#=======================================================================
#    $Id: Template.pm,v 1.5 2008/05/12 12:18:09 pythontech Exp $
#	Expand templates
#
#    Copyright (C) 2000-2008  Colin Hogben, Python Technology Ltd.
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
#	Based on ideas of WebMacro and MetaText
#	WebMacro doesn't do Perl (Java only).
#	MetaText doesn't do loops.
#	Not sure that either do escaping very conveniently.
#	This is very loose in its checking.
#
#	%%IF var%% ... %%ENDIF%%		Conditional
#	%%UNLESS var%% ... %%ENDIF%%		 "
#	%%LOOP var%% ... %%ENDLOOP%%		Loop over array of hashes
#	%%BLOCK var%% ... %%ENDBLOCK%%		Expand now and set var
#	%%DEFINE name%% ... %%ENDDEF%%		Macro
#	%%INCLUDE file%%
#	%%SET var=value%%
#	%%var%%
#=======================================================================
package PythonTech::Template;
use strict;
use Carp;

#-----------------------------------------------------------------------
#	Attributes:
#	DIRS	directory list for finding templates (via INCLUDE etc.)
#		can be array ref, or colon-separated string
#	STAG	string indicating start of tag, e.g. "[["
#	ETAG	string indicating end of tag, e.g. "]]"
#	COOKED	name of method for getting formatted value, e.g. "HTML"
#	RAW	name of method for getting unformatted value, e.g. "GET"
#	ESCAPE	code reference for formatting value
#	CONTEXT	error message prefix
#	USERDATA	Exta value passed in to RAW/COOKED methods
#-----------------------------------------------------------------------
sub new {
    my $class = shift;
    my $self = {
	DIRS => ".",
	ESCAPE => \&html_quote,
	STAG => undef,
	ETAG => undef,
	COOKED => 'HTML',
	RAW => 'GET',
	CONTEXT => '',
	USERDATA => undef,
    };
    bless $self, $class;
    if (ref($_[0]) eq 'HASH') {
	$self->configure(%{$_[0]});
    } else {
	$self->configure(@_);
    }
    return $self;
}

sub configure {
    my $self = shift;
    while (my($key,$value) = splice(@_,0,2)) {
	die ref($self).": Unknown attribute $key"
	    unless exists $self->{$key};
	$self->{$key} = $value;
    }
}

#-----------------------------------------------------------------------
#	Lexical analysis: split on %% gubbins... %%
#	Note that newline following closing %% is gobbled, if present.
#	Consumed by 'blockish' functions; possibly put back by others.
#-----------------------------------------------------------------------
sub _lex {
    my $self = shift;
    my($text) = @_;
    my $stag = $self->{'STAG'};
    my $etag = $self->{'ETAG'};
    if ($stag eq '') {
	# Guess tags from initial part of file
	($stag) = $text =~ /^\s*([^\s\w\#]+)/;
	($etag = reverse $stag) =~ tr/\<\[\(\{/\>\]\)\}/;
	$self->{'STAG'} = $stag;
	$self->{'ETAG'} = $etag;
    }
    # Directives are bounded by STAG and ETAG
    # Within that, there may be strings quoted with "".
    # Within such a string there may be ETAG, or characters escaped with \ .
    my @tokens = split(/(\Q$stag\E[\w\#](?:[^\"]|\"(?:\\.|[^\"\\]+)*\")*?\Q$etag\E\n?)/,$text);
    return @tokens;
}

#-----------------------------------------------------------------------
#	Recursively parse the token stream to produce tree.
#	Passed a list of end tokens to stop at; if it the first.
#	Returns one of these tree nodes:
#	  [SEQ, args, item...]		item is text or a node
#	  [IF, args, thenbody, elsebody]
#	  [UNLESS, args, body]
#	  [LOOP, args, body]
#	  [WITH, args, body]
#	  [DEFINE, args, body]
#	  [BLOCK, args, body]
#	  [<other>, args, nl]
#-----------------------------------------------------------------------
sub _parse {
    my $self = shift;
    my($src,$pline,$tokens,$ends) = @_;
    my $stag = $self->{'STAG'};
    my $etag = $self->{'ETAG'};
    my @stack = ();
    my @seq = ("$src:$$pline",'SEQ');
    while (@$tokens) {
	my $t = shift @$tokens;
	if ((my $func, my $args, my $nl) = 
	    ($t =~ /^$stag\s*(\S+)\s*(.*?)\s*$etag(\n?)$/)) {
#	    print "func:$func args=$args\n";
	    my $where = "$src:$$pline";
	    $$pline += 1 if $nl;

	    if (grep {$_ eq $func} @$ends) {
		# Have found end token
		my $ret = \@seq;
		return wantarray ? ($ret, $func, $args) : $ret;
	    } elsif ($func eq 'IF' || $func eq 'UNLESS') {
		my $cond = $args;
		my @cases;
		my $else;
		my($then,$etok,$econd) = $self->_parse($src,$pline,$tokens,
						       ['ELSEIF', 'ELSE', 'ENDIF']);
		push @cases, [$where, $func, $cond, $then];
		while ($etok eq 'ELSEIF') {
		    $where = "$src:$$pline";
		    $cond = $econd;
		    ($then,$etok,$econd) = $self->_parse($src,$pline,$tokens,
							 ['ELSEIF', 'ELSE', 'ENDIF']);
		    push @cases, [$where, 'IF', $cond, $then];
		}
		if ($etok eq 'ELSE') {
		    $else = $self->_parse($src,$pline,$tokens, ['ENDIF']);
		}
		my $exp = $else;
		while (@cases) {
		    my $case = pop @cases;
		    $exp = [@$case, $exp];
		}
		push(@seq, $exp);
	    } elsif ($func eq 'LOOP') {
		my $loop = $self->_parse($src,$pline,$tokens, ['ENDLOOP']);
		push(@seq, [$where, 'LOOP', $args, $loop]);
	    } elsif ($func eq 'WITH') {
		my $body = $self->_parse($src,$pline,$tokens, ['ENDWITH']);
		push(@seq, [$where, 'WITH', $args, $body]);
	    } elsif ($func eq 'BLOCK') {
		my $blk = $self->_parse($src,$pline,$tokens, ['ENDBLOCK']);
		push(@seq, [$where, 'BLOCK', $args, $blk]);
	    } elsif ($func eq 'DEFINE') {
		my $blk = $self->_parse($src,$pline,$tokens, ['ENDDEF']);
		push(@seq, [$where, 'DEFINE', $args, $blk]);
	    } elsif ($func =~ /^(END(IF|LOOP|WITH|DEF|BLOCK)|ELSEIF)$/) {
		die "$where: $self->{'CONTEXT'}Misplaced $func\n";
	    } else {
		# INCLUDE, SET, LENGTH, VALUE, var, ...
		push(@seq, [$where, $func, $args, $nl]);
	    }
	} else {
	    # Literal text
	    push(@seq, $t) unless $t eq '';
	    my @nl = $t =~ m!\n!g;
	    $$pline += scalar @nl;
#	    print "text:$t\n";
	}
    }
    # Reached end of tokenlist
    if (@$ends) {
	my $choice = join ", ", @$ends;
	die "$src:$$pline: $self->{'CONTEXT'}End of template while expecting $choice\n";
    } else {
	return \@seq;
    }
}

#-----------------------------------------------------------------------
#	For debugging: dump to STDOUT
#-----------------------------------------------------------------------
sub dump {
    my $self = shift;
    my($level,@tree) = @_;
    my $where = shift @tree;
    my $type = shift @tree;
    if ($type eq 'SEQ') {
	print $where,"  "x$level,'SEQ',"\n";
	foreach my $t (@tree) {
	    if (ref($t) eq 'ARRAY') {
		$self->dump($level+1,@$t);
	    } else {
		$t =~ s/\n/\\n/g;
		print "  "x($level+1),"\"$t\"\n";
	    }
	}
    } elsif ($type =~ /^(IF|UNLESS|LOOP|WITH|BLOCK|DEFINE)/) {
	print $where,"  "x$level,"$type $tree[0]\n";
	$self->dump($level+1,@{$tree[1]});
	if (($type eq 'IF' || $type eq 'UNLESS') && defined $tree[2]) {
	    print "  "x$level,"ELSE\n";
	    $self->dump($level+1,@{$tree[2]});
	}
    } else {
	print $where,"  "x$level,"$type $tree[0]\n";
    }
}

#-----------------------------------------------------------------------
#	Generate output text from parse tree and supplied var lookup
#	Inputs:
#	  $tree		Parse tree
#	  $local	Local variables in current scope
#	  @varsets	Environment for non-local lookups
#	The reason for passing in $local is that SEQ calls _tree2cb
#	recursively, so we want the same variables visible.
#	A new scope is started:
#	  when including a file;	(is this desirable?)
#	  in a WITH or LOOP
#-----------------------------------------------------------------------
sub _tree2text {
    my($self, $tree,$local,@varsets) = @_;
    my $out = "";
    $self->_tree2cb(sub {$out .= $_[0]},
		    $tree,$local,@varsets);
    return $out;
}

sub _tree2cb {
    my($self, $cb,$tree,$local,@varsets) = @_;
    my($where,$func,@args) = @$tree;
    if ($func eq 'SEQ') {
	foreach my $t (@args) {
	    if (ref($t) eq 'ARRAY') {
		$self->_tree2cb($cb,$t,$local,@varsets);
	    } else {
		$cb->($t);
	    }
	}
    } elsif ($func eq 'IF' || $func eq 'UNLESS') {
	my($test,$then,$else) = @args;
	my $val = $self->_get($where,$test,undef, $local,@varsets);
	# Suppress for normal Perl 'FALSE' values, or empty array
	my $test = ($val && !(ref($val) eq 'ARRAY' && @$val == 0));
	$test = ! $test if $func eq 'UNLESS';
	if ($test) {
	    $self->_tree2cb($cb,$then,$local,@varsets);
	} elsif (defined $else) {
	    $self->_tree2cb($cb,$else,$local,@varsets);
	}
    } elsif ($func eq 'LOOP') {
	my($var,$body) = @args;
	my $val = $self->_get($where,$var,undef, $local,@varsets);
	if (!defined($val)) {
	    warn "$where: $self->{'CONTEXT'}Loop variable `$var' is not defined\n";
	} elsif (ref($val) ne 'ARRAY') {
	    warn "$where: $self->{'CONTEXT'}`$var' is ".ref($val)." not array (value=$val)\n";
	} else {
	    foreach my $elem (@$val) {
		if (! &_is_hash_or_obj($elem)) {
		    die "$where: $self->{'CONTEXT'}Unexpected ". (ref($elem) || "scalar $elem")." in $var loop\n";
		} else {
		    $self->_tree2cb($cb,$body, {},$elem,$local,@varsets);
		}
	    }
	}
    } elsif ($func eq 'WITH') {
	my($var,$body) = @args;
	my $val = $self->_get($where,$var,undef, $local,@varsets);
	if (!defined($val)) {
	    warn "$where: $self->{'CONTEXT'}`WITH' variable `$var' is not defined\n";
	} elsif (! &_is_hash_or_obj($val)) {
	    die "$where: $self->{'CONTEXT'}Unexpected ". (ref($val) || "scalar $val")." in WITH\n";
	} else {
	    $self->_tree2cb($cb,$body, {},$val,$local,@varsets);
	}
    } elsif ($func eq 'BLOCK') {
	my($name,$body) = @args;
	my $block = "";
	$self->_tree2cb(sub {$block .= $_[0]},
			$body, {},$local,@varsets);
	$local->{"$self->{'COOKED'}:$name"} = $block;
#    } elsif ($func eq 'DEFINE') {
#	my($name,$body) = @args;
#	$varsets[0]->{$name} = [\&_macro, $self, $body];
    } elsif ($func eq 'INCLUDE') {
	my($file) = @args;
	my $text = $self->_readfile('INC'.$where,$file);
	$self->_tmpl2cb($cb,$file,$text, $local,@varsets);
    } elsif ($func eq 'SET') {
	my($vv) = @args;
	if ((my $var,my $value) = split(/=/,$vv,2)) {
	    $local->{$var} = $value;
	}
    } elsif ($func eq 'LENGTH') {
	my($var) = @args;
	my $val = $self->_get($where,$var,undef, $local,@varsets);
	if (!defined($val)) {
	    warn "$where: $self->{'CONTEXT'}Length of unknown `$var'\n";
	} else {
	    if (ref($val) eq 'ARRAY') {
		$cb->(scalar @$val);
	    } elsif (ref($val) eq 'HASH') {
		$cb->(scalar keys %$val);
	    } elsif (ref($val) eq '') {
		$cb->(length $val);
	    } else {
		warn "$where: $self->{'CONTEXT'}Cannot get length of ".ref($val)."\n";
	    }
	}
    } elsif ($func eq 'TEMPLATEINFO') {
	$cb->("DO NOT EDIT THIS FILE! It was generated from a template.");
    } elsif ($func eq 'PLURAL') {
	my($opts,$nl) = @args;
	my($var,$plural,$single) = split(/\s+/,$opts);
	my $val = $self->_get($where,$var,undef, $local,@varsets);
	my $count;
	if (!defined($val)) {
	    $count = 0;
	} else {
	    if (ref($val) eq 'ARRAY') {
		$count = scalar @$val;
	    } elsif (ref($val) eq 'HASH') {
		$count = scalar keys %$val;
	    } else {
		$count = 0 + $val;
	    }
	}
	if ($count >= 2) {
	    $cb->(defined($plural) ? $plural : "s");
	} else {
	    $cb->($single) if defined($single);
	}
    } elsif ($func eq 'VALUE') {
	my($var,$body) = @args;
	my($val,$cooked) = $self->_get($where,$var,1, $local,@varsets);
	if (! defined($val)) {
	    warn "$where: $self->{'CONTEXT'}No variable `$var' defined\n";
	} elsif (! $cooked) {
	    my $esc = $self->{'ESCAPE'};
	    $cb->(&{$esc}($val));
	} else {
	    $cb->($val);
	}
    } elsif ($func =~ /^\#/) {
	;			# Comment - ignore
    } else {
	my($opts,$nl) = @args;	# opts unused
	my($val,$cooked) = $self->_get($where,$func,1, $local,@varsets);
	if (! defined($val)) {
	    warn "$where: $self->{'CONTEXT'}No variable `$func' defined\n";
	} elsif (! $cooked) {
	    my $esc = $self->{'ESCAPE'};
	    $cb->(&{$esc}($val));
	} else {
	    $cb->($val);
	}
    }
}

#-----------------------------------------------------------------------
#	Fetch an entity out of a list of hashes (or objects)
#	Try each in turn until a defined value is found:
#	If it is an object:
#	* If it has a method as named by the COOKED attribute of the
#	  Template, call that with the varname as a single argument.
#	* Else if it has a method as named by RAW attribute, call that
#	  and then apply the ESCAPE function (if it is a scalar)
#	If it is an object with neither of these methods, or it is a
#	hash, check if it has a member $varname (treated as cooked);
#	or a member _$varname (uncooked)
#	Return $value, or ($value, 1) in an array context if it has
#	been cooked already.
#-----------------------------------------------------------------------
sub _get1 {
    my $self = shift;
    my($where,$varname,$wantcooked, @envlist) = @_;
    foreach my $thing (@envlist) {
	die "$where: $self->{'CONTEXT'}Unexpected ".&_desc($thing)." in envlist\n"
	    unless &_is_hash_or_obj($thing);
	if (ref($thing) ne 'HASH') {
	    # Have a blessed object.
	    # It should implement the 'RAW' method (e.g. GET).
	    # It may implement the 'COOKED' method (e.g. HTML).
#	    print "can on ",ref($thing)," $thing\n";
	    my $userdata = $self->{'USERDATA'};
	    # See if the object has a COOKED (e.g. HTML) method
	    my $cooked = $thing->can($self->{'COOKED'});
	    if (defined($cooked)) {
		my $val = &$cooked($thing, $varname, $userdata);
		return wantarray ? ($val, 1) : $val
		    if defined($val);
	    }
	    my $raw = $thing->can($self->{'RAW'});
	    if (defined($raw)) {
		my $val = &$raw($thing, $varname, $userdata);
		return $val if defined $val;
	    }
	    # If has no raw/cooked method, try direct lookup
	    # by falling through to code below for hash refs.
	    #
	    # But this will fail if $thing is a blessed
	    # array/scalar/subr ref, hence test with eval.
	    # Will also return false for an empty hash, but that's
	    # OK as we want to return undef in that case too.
	    # FIXME: Is there a better way to test what was blessed?
	    # FIXME: The test gives false positive for blessed array
	    # FIXME: should we really be poking into the internals at all?

	    next unless eval {%$thing};
	} else {
	    # Just have a plain hash.
	    # Look for a raw member (e.g. 'foo') or a cooked member (e.g. 'HTML:foo')
	    if (defined(my $val = $thing->{$varname})) {
		return $val;
	    }
	    if (defined(my $val = $thing->{"$self->{'COOKED'}:$varname"})) {
		return wantarray ? ($val, 1) : $val;
	    }
	}
    }
    return undef;
}

sub _get {
    my($self, $where,$varpath,$wantcooked,@envlist) = @_;
    my @path = split m!/!, $varpath;
    # Find first step from all env
    my $step1 = shift @path;
    my($val,$cooked) = $self->_get1($where,$step1,$wantcooked, @envlist);
    # After each step lookup, we have either
    # - an object
    # - a cooked string
    # - a raw string
    # - an array
    # Only an object can be further stepped into, but all may be returned
    while (my $step = shift @path) {
	if (! &_is_hash_or_obj($val)) {
	    die "$where: $self->{'CONTEXT'}".
		"Cannot extract subpart $step from ".&_desc($val)."\n";
	} else {
	    ($val,$cooked) = $self->_get1($where,$step,$wantcooked, $val);
	}
    }
    return wantarray ? ($val,$cooked) : $val;
}

sub _subst {
    my $self = shift;
    my($where,$vars,$val,$args,$nl) = @_;
    if (ref($val) eq 'ARRAY') {
	my($func,@rest) = @$val;
	if (ref($func) eq 'CODE') {
	    # Call routine with args given, appended with var hash and args
	    my $result = &{$func}(@rest, $vars, $args);
	    return $result;
	} else {
	    warn "$where: $self->{'CONTEXT'}Substitution of a non-code array";
	}
    } else {
	return $val;
    }
}

sub _macro {
    my $self = shift;
    my($body,$vars,$args) = @_;
    my %macvars = (%$vars);
    my $mac = $self->_tree2text(\%macvars, $body);
    return $mac;
}

#-----------------------------------------------------------------------
#	Test if a value is either a HASH or a blessed reference.
#	Perl guarantees that built-in refs are all uppercase.
#-----------------------------------------------------------------------
sub _is_hash_or_obj {
    my $thing = shift;
    return 1 if ref($thing) =~ /[^A-Z]/; # Blessed
    return (ref($thing) eq 'HASH');
}

sub _desc {
    my $thing = shift;
    if (! defined $thing) {
	return "undefined value";
    } elsif (ref($thing) eq '') {
	return "scalar value";
    } elsif (ref($thing) eq 'ARRAY') {
	return "array value";
    } else {
	return "value";
    }
}
    
sub _noop {
#    print "_noop $_[0]";
    return shift;
}

sub _readfile {
    my $self = shift;
    my($whence,$rel) = @_;
    my $file;
    if ($rel =~ m|^(\.\.?)?/|) { # Relative to current or parent dir or /
	$file = $rel;
    } else {
	my @dirs;
	if (ref($self->{'DIRS'}) eq 'ARRAY') {
	    @dirs = @{$self->{'DIRS'}};
	} else {
	    @dirs = split(/:/, $self->{'DIRS'});
	}
	foreach my $dir (@dirs) {
	    my $abs = "$dir/$rel";
	    ($file = $abs, last) if -f $abs && -r $abs;
	}
	die "$whence: No such file: $rel\n" unless defined $file;
    }
    local(*FILE,$/);
    open(FILE,'<',$file) || die "Cannot open $file: $!\n";
    my $text = <FILE>;
    close(FILE);
    return $text;
}

sub dump_file {
    my $self = shift;
    my($file,@varsets) = @_;
    my $text = $self->_readfile("",$file);
    my @toks = $self->_lex($text);
    my $line = 1;
    my $tree = $self->_parse($file,\$line,\@toks, []);
    $self->dump(0,@$tree);
}

sub _tmpl2text {
    my($self, $src,$text,@varsets) = @_;
    my $out = "";
    $self->_tmpl2cb(sub {$out .= $_[0]},
		    $src,$text,@varsets);
    return $out;
}

sub _tmpl2cb {
    my($self, $cb,$src,$text,@varsets) = @_;
    my @toks = $self->_lex($text);
    my $line = 1;
    my $tree = $self->_parse($src,\$line,\@toks, []);
    $self->_tree2cb($cb,$tree,{},@varsets);
}

sub process_file {
    my $self = shift;
    my($file,@varsets) = @_;
    my $text = $self->_readfile("",$file);
    my $out = $self->_tmpl2text($file,$text,@varsets);
    return $out;
}

sub process_text {
    my $self = shift;
    my($text,@varsets) = @_;
    my $out = $self->_tmpl2text("<inline>",$text,@varsets);
    return $out;
}

sub html_quote {
    my($text) = @_;
    $text =~ s/&/&amp;/g;
    $text =~ s/</&lt;/g;
    $text =~ s/>/&gt;/g;
    $text =~ s/\"/&quot;/g;
    $text =~ s/\015//g;
    $text =~ s/([^\012\040-\176])/sprintf("\&#%d;",unpack("C",$1))/eg;
    return $text;
}

#=======================================================================
#	Macro expander (unused)
#=======================================================================
package PythonTech::Template::Macro;

sub new {
    my $class = shift;
    my($tmpl,$body,$cooked) = @_;
    my $self = {tmpl => $tmpl, body => $body};
    bless $self, $class;
    return $self;
}

sub GET {
    my $self = shift;
    my($varname) = @_;
}

1;
